package be.iminds.ilabt.jfed.util.lib;

import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.ssh_terminal_tool.known_hosts.OpenSshKnownHostsFile;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import be.iminds.ilabt.jfed.util.library.OpenSSHKnownHostsInMemory;
import be.iminds.ilabt.jfed.util.library.SshProxySocketImplFactory;
import com.hierynomus.sshj.signature.SignatureEdDSA;
import net.schmizz.sshj.DefaultConfig;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.channel.direct.LocalPortForwarder;
import net.schmizz.sshj.signature.SignatureDSA;
import net.schmizz.sshj.signature.SignatureRSA;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.userauth.UserAuthException;
import net.schmizz.sshj.userauth.keyprovider.KeyPairWrapper;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Easily forward local ports to remote machines, over an SSH proxy
 *
 * This caches so that a request to the same remoteHostName:remotePort returns the same
 */
@Singleton
public class LocalForwardUtil {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(LocalForwardUtil.class);

    private final GeniUserProvider geniUserProvider;
    private final OpenSshKnownHostsFile openSshKnownHostsFile;

    private static final int START_FORWARD_PORT = 11111;
    private int nextForwardPort = START_FORWARD_PORT;

    private Map<ForwardKey, Forward> activeForwards = new ConcurrentHashMap<>();

    @Inject
    public LocalForwardUtil(GeniUserProvider geniUserProvider, OpenSshKnownHostsFile openSshKnownHostsFile) {
        this.geniUserProvider = geniUserProvider;
        this.openSshKnownHostsFile = openSshKnownHostsFile;
    }

    /**
     * Forward a local port to remote machines, over an SSH proxy
     * @param sshProxyInfo the SSH proxy to use
     * @param remoteHostName the hostname of the target remote machine
     * @param remotePort the port of the target remote machine
     * @return the local port that is forwarded to remoteHostName:remotePort
     */
    public int forwardOverSshProxy(@Nonnull JFedConnection.SshProxyInfo sshProxyInfo,
                                   @Nonnull String remoteHostName,
                                   int remotePort) throws IOException {
        final ForwardKey forwardKey = new ForwardKey(sshProxyInfo, remoteHostName, remotePort);
        Forward foundForward = activeForwards.get(forwardKey);
        if (foundForward != null) {
            LOG.debug("Found existing LocalForward: "+foundForward);
            return foundForward.getLocalPort();
        }



        //no port forwarind yet -> Setup port forward and do bookkeeping
        int forwardPort = nextForwardPort++;

        DefaultConfig config = new DefaultConfig() {
            @Override
            protected void initSignatureFactories() {
                setSignatureFactories(new SignatureRSA.Factory(), new SignatureDSA.Factory(), new SignatureEdDSA.Factory());

            }
        };
        SSHClient sshClient = new SSHClient(config);

        if (sshProxyInfo.getHostKey() != null) {
            sshClient.addHostKeyVerifier(new OpenSSHKnownHostsInMemory(sshProxyInfo.getHostKey()));
        } else {
            LOG.warn("Cannot verify proxy host as no known_hosts-line is specified");
            sshClient.addHostKeyVerifier(new PromiscuousVerifier());
        }

        sshClient.setTimeout(SshProxySocketImplFactory.KEX_TIMEOUT_MS);
        sshClient.setConnectTimeout(SshProxySocketImplFactory.CONNECT_TIMEOUT_MS);

        LOG.debug("Connecting to " + sshProxyInfo.getHostname() + ":" + sshProxyInfo.getPort() +
                " as user "+sshProxyInfo.getUsername()+" with timeout " + SshProxySocketImplFactory.CONNECT_TIMEOUT_MS+" ms");
        sshClient.connect(sshProxyInfo.getHostname(), sshProxyInfo.getPort());

        LOG.debug("Connect done. Will now authenticate.");

        try {
            assert sshProxyInfo.getUsername() != null;
            assert sshProxyInfo.getSshKeyInfo() != null;
            assert sshProxyInfo.getSshKeyInfo().getPrivateKey() != null;
            assert KeyUtil.matchingKeys(sshProxyInfo.getSshKeyInfo().getPublicKey(), sshProxyInfo.getSshKeyInfo().getPrivateKey());

            sshClient.authPublickey(sshProxyInfo.getUsername(),
                    new KeyPairWrapper(sshProxyInfo.getSshKeyInfo().getPublicKey(), sshProxyInfo.getSshKeyInfo().getPrivateKey()));
            LOG.debug("Connected and authenticated to {}:{}", sshProxyInfo.getHostname(), sshProxyInfo.getPort());

        } catch (UserAuthException ex) {
            LOG.warn("Connected to {}:{}, but authentication failed", sshProxyInfo.getHostname(), sshProxyInfo.getPort());
            throw new IOException("Failed to authenticate user \"" + sshProxyInfo.getUsername() + "\"" +
                    " on server " + sshProxyInfo.getHostname() + ":" + sshProxyInfo.getPort() + "", ex);
        }

        final LocalPortForwarder.Parameters params
                = new LocalPortForwarder.Parameters("0.0.0.0", forwardPort,
                remoteHostName, remotePort);
        final ServerSocket ss = new ServerSocket();
        ss.setReuseAddress(true);
        ss.bind(new InetSocketAddress(params.getLocalHost(), params.getLocalPort()));
        LocalPortForwarder localPortForwarder = sshClient.newLocalPortForwarder(params, ss);
        Thread listenThread = new Thread(() -> {
            try {
                localPortForwarder.listen();
            } catch (IOException e) {
                LOG.warn("Error while forwarding "+forwardKey, e);
                close(forwardKey);
            } finally {
                LOG.warn("Closed forwarding "+forwardKey);
                close(forwardKey);
            }
        });
        Forward forward = new Forward(
                sshProxyInfo, localPortForwarder, remoteHostName, remotePort, forwardPort, sshClient, listenThread
        );
        listenThread.start();
        assert forward.getKey().equals(forwardKey);
        activeForwards.put(forwardKey, forward);

        return forwardPort;
    }


    private void close(@Nullable ForwardKey forwardKey) {
        if (forwardKey == null)
            return;

        Forward foundForward = activeForwards.remove(forwardKey);
        if (foundForward != null) {
            foundForward.close();
        }
    }

    public void close(JFedConnection.SshProxyInfo sshProxyInfo, String remoteHostName, int remotePort) {
        close(new ForwardKey(sshProxyInfo, remoteHostName, remotePort));
    }


    public void close(int localPort) {
        //TODO specific SSH connection
        throw new RuntimeException("Not yet implemented");
    }

    public void close() {
        //TODO call this where appropriate

        Set<ForwardKey> allKeys = new HashSet<>(activeForwards.keySet());
        for (ForwardKey key : allKeys) {
            close(key);
        }
    }


    private static class ForwardKey {
        @Nonnull private final JFedConnection.SshProxyInfo sshProxyInfo;
        @Nonnull private final String remoteHostName;
        private final int remotePort;

        public ForwardKey(@Nonnull JFedConnection.SshProxyInfo sshProxyInfo, @Nonnull String remoteHostName, int remotePort) {
            this.sshProxyInfo = sshProxyInfo;
            this.remoteHostName = remoteHostName;
            this.remotePort = remotePort;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ForwardKey forward = (ForwardKey) o;

            if (remotePort != forward.remotePort) return false;
            if (!sshProxyInfo.equals(forward.sshProxyInfo)) return false;
            return remoteHostName.equals(forward.remoteHostName);
        }

        @Override
        public int hashCode() {
            int result = sshProxyInfo.hashCode();
            result = 31 * result + remoteHostName.hashCode();
            result = 31 * result + remotePort;
            return result;
        }
    }
    private static class Forward {
        @Nonnull private final JFedConnection.SshProxyInfo sshProxyInfo;
        @Nullable private LocalPortForwarder localPortForwarder;
        @Nullable private SSHClient sshClient;
        @Nonnull private final String remoteHostName;
        private final int remotePort;
        private final int localPort;
        @Nonnull private final Thread listenThread;

        public Forward(@Nonnull JFedConnection.SshProxyInfo sshProxyInfo,
                       @Nullable LocalPortForwarder localPortForwarder,
                       @Nonnull String remoteHostName,
                       int remotePort,
                       int localPort,
                       @Nullable SSHClient sshClient,
                       @Nonnull Thread listenThread) {
            this.sshProxyInfo = sshProxyInfo;
            this.localPortForwarder = localPortForwarder;
            this.sshClient = sshClient;
            this.remoteHostName = remoteHostName;
            this.remotePort = remotePort;
            this.localPort = localPort;
            this.listenThread = listenThread;
        }

        public ForwardKey getKey() {
            return new ForwardKey(sshProxyInfo, remoteHostName, remotePort);
        }

        public int getLocalPort() {
            return localPort;
        }

        public void close() {
            try {
                if (localPortForwarder != null) {
                    try {
                        localPortForwarder.close();
                    } finally {
                        localPortForwarder = null;
                    }
                }
                if (sshClient != null) {
                    try {
                        sshClient.close();
                    } finally {
                        sshClient = null;
                    }
                }
            } catch (IOException e) {
                //ignore exception while closing
            }
        }

        @Override
        public String toString() {
            return "Forward{" +
                    "sshProxyInfo=" + sshProxyInfo +
                    ", localPortForwarder=" + localPortForwarder +
                    ", sshClient=" + (sshClient!=null ? "non-null" : "null") +
                    ", remoteHostName='" + remoteHostName + '\'' +
                    ", remotePort=" + remotePort +
                    ", localPort=" + localPort +
                    '}';
        }
    }
}
